"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const MysqlSchemaBuilder_1 = require("../schema-builder/MysqlSchemaBuilder");
const ConnectionIsNotSetError_1 = require("./error/ConnectionIsNotSetError");
const DriverPackageNotInstalledError_1 = require("./error/DriverPackageNotInstalledError");
const DriverPackageLoadError_1 = require("./error/DriverPackageLoadError");
const DriverUtils_1 = require("./DriverUtils");
const ColumnTypes_1 = require("../metadata/types/ColumnTypes");
const TransactionNotStartedError_1 = require("./error/TransactionNotStartedError");
const TransactionAlreadyStartedError_1 = require("./error/TransactionAlreadyStartedError");
/**
 * This driver organizes work with mysql database.
 */
class MysqlDriver {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(options, logger, mysql) {
        /**
         * Pool of database connections.
         */
        this.databaseConnectionPool = [];
        this.options = DriverUtils_1.DriverUtils.buildDriverOptions(options);
        this.logger = logger;
        this.mysql = mysql;
        // validate options to make sure everything is set
        DriverUtils_1.DriverUtils.validateDriverOptions(this.options);
        // if mysql package instance was not set explicitly then try to load it
        if (!mysql)
            this.loadDependencies();
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Performs connection to the database.
     * Based on pooling options, it can either create connection immediately,
     * either create a pool and create connection when needed.
     */
    connect() {
        // build connection options for the driver
        const options = Object.assign({}, {
            host: this.options.host,
            user: this.options.username,
            password: this.options.password,
            database: this.options.database,
            port: this.options.port
        }, this.options.extra || {});
        // pooling is enabled either when its set explicitly to true,
        // either when its not defined at all (e.g. enabled by default)
        if (this.options.usePool === undefined || this.options.usePool === true) {
            this.pool = this.mysql.createPool(options);
            return Promise.resolve();
        }
        else {
            return new Promise((ok, fail) => {
                this.databaseConnection = {
                    id: 1,
                    connection: this.mysql.createConnection(options),
                    isTransactionActive: false
                };
                this.databaseConnection.connection.connect((err) => err ? fail(err) : ok());
            });
        }
    }
    /**
     * Closes connection with database.
     */
    disconnect() {
        if (!this.databaseConnection && !this.pool)
            throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
        return new Promise((ok, fail) => {
            const handler = (err) => err ? fail(err) : ok();
            // if pooling is used, then disconnect from it
            if (this.pool) {
                this.pool.end(handler);
                this.pool = undefined;
                this.databaseConnectionPool = [];
            }
            // if single connection is opened, then close it
            if (this.databaseConnection) {
                this.databaseConnection.connection.end(handler);
                this.databaseConnection = undefined;
            }
        });
    }
    /**
     * Retrieves a new database connection.
     * If pooling is enabled then connection from the pool will be retrieved.
     * Otherwise active connection will be returned.
     */
    retrieveDatabaseConnection() {
        if (this.pool) {
            return new Promise((ok, fail) => {
                this.pool.getConnection((err, connection) => {
                    if (err)
                        return fail(err);
                    let dbConnection = this.databaseConnectionPool.find(dbConnection => dbConnection.connection === connection);
                    if (!dbConnection) {
                        dbConnection = {
                            id: this.databaseConnectionPool.length,
                            connection: connection,
                            isTransactionActive: false
                        };
                        this.databaseConnectionPool.push(dbConnection);
                    }
                    ok(dbConnection);
                });
            });
        }
        if (this.databaseConnection)
            return Promise.resolve(this.databaseConnection);
        throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
    }
    /**
     * Releases database connection. This is needed when using connection pooling.
     * If connection is not from a pool, it should not be released.
     */
    releaseDatabaseConnection(dbConnection) {
        if (this.pool) {
            dbConnection.connection.release();
            this.databaseConnectionPool.splice(this.databaseConnectionPool.indexOf(dbConnection), 1);
        }
        return Promise.resolve();
    }
    /**
     * Creates a schema builder which can be used to build database/table schemas.
     */
    createSchemaBuilder(dbConnection) {
        return new MysqlSchemaBuilder_1.MysqlSchemaBuilder(this, dbConnection);
    }
    /**
     * Removes all tables from the currently connected database.
     */
    clearDatabase(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
            const disableForeignKeysCheckQuery = `SET FOREIGN_KEY_CHECKS = 0;`;
            const dropTablesQuery = `SELECT concat('DROP TABLE IF EXISTS ', table_name, ';') AS query FROM information_schema.tables WHERE table_schema = '${this.options.database}'`;
            const enableForeignKeysCheckQuery = `SET FOREIGN_KEY_CHECKS = 1;`;
            yield this.query(dbConnection, disableForeignKeysCheckQuery);
            const dropQueries = yield this.query(dbConnection, dropTablesQuery);
            yield Promise.all(dropQueries.map(query => this.query(dbConnection, query["query"])));
            yield this.query(dbConnection, enableForeignKeysCheckQuery);
        });
    }
    /**
     * Starts transaction.
     */
    beginTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (dbConnection.isTransactionActive)
                throw new TransactionAlreadyStartedError_1.TransactionAlreadyStartedError();
            yield this.query(dbConnection, "START TRANSACTION");
            dbConnection.isTransactionActive = true;
        });
    }
    /**
     * Commits transaction.
     */
    commitTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new TransactionNotStartedError_1.TransactionNotStartedError();
            yield this.query(dbConnection, "COMMIT");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Rollbacks transaction.
     */
    rollbackTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new TransactionNotStartedError_1.TransactionNotStartedError();
            yield this.query(dbConnection, "ROLLBACK");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Executes a given SQL query.
     */
    query(dbConnection, query, parameters) {
        if (!this.databaseConnection && !this.pool)
            throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
        this.logger.logQuery(query);
        return new Promise((ok, fail) => {
            dbConnection.connection.query(query, parameters, (err, result) => {
                if (err) {
                    this.logger.logFailedQuery(query);
                    this.logger.logQueryError(err);
                    return fail(err);
                }
                ok(result);
            });
        });
    }
    /**
     * Insert a new row with given values into given table.
     */
    insert(dbConnection, tableName, keyValues, idColumnName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
            const keys = Object.keys(keyValues);
            const columns = keys.map(key => this.escapeColumnName(key)).join(", ");
            const values = keys.map(key => "?").join(",");
            const parameters = keys.map(key => keyValues[key]);
            const sql = `INSERT INTO ${this.escapeTableName(tableName)}(${columns}) VALUES (${values})`;
            const result = yield this.query(dbConnection, sql, parameters);
            return result.insertId;
        });
    }
    /**
     * Updates rows that match given conditions in the given table.
     */
    update(dbConnection, tableName, valuesMap, conditions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
            const updateValues = this.parametrize(valuesMap).join(", ");
            const conditionString = this.parametrize(conditions).join(" AND ");
            const sql = `UPDATE ${this.escapeTableName(tableName)} SET ${updateValues} ${conditionString ? (" WHERE " + conditionString) : ""}`;
            const conditionParams = Object.keys(conditions).map(key => conditions[key]);
            const updateParams = Object.keys(valuesMap).map(key => valuesMap[key]);
            const allParameters = updateParams.concat(conditionParams);
            yield this.query(dbConnection, sql, allParameters);
        });
    }
    /**
     * Deletes from the given table by a given conditions.
     */
    delete(dbConnection, tableName, conditions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
            const conditionString = this.parametrize(conditions).join(" AND ");
            const sql = `DELETE FROM ${this.escapeTableName(tableName)} WHERE ${conditionString}`;
            const parameters = Object.keys(conditions).map(key => conditions[key]);
            yield this.query(dbConnection, sql, parameters);
        });
    }
    /**
     * Inserts rows into the closure table.
     */
    insertIntoClosureTable(dbConnection, tableName, newEntityId, parentId, hasLevel) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("mysql");
            let sql = "";
            if (hasLevel) {
                sql = `INSERT INTO ${this.escapeTableName(tableName)}(ancestor, descendant, level) ` +
                    `SELECT ancestor, ${newEntityId}, level + 1 FROM ${this.escapeTableName(tableName)} WHERE descendant = ${parentId} ` +
                    `UNION ALL SELECT ${newEntityId}, ${newEntityId}, 1`;
            }
            else {
                sql = `INSERT INTO ${this.escapeTableName(tableName)}(ancestor, descendant) ` +
                    `SELECT ancestor, ${newEntityId} FROM ${this.escapeTableName(tableName)} WHERE descendant = ${parentId} ` +
                    `UNION ALL SELECT ${newEntityId}, ${newEntityId}`;
            }
            yield this.query(dbConnection, sql);
            const results = yield this.query(dbConnection, `SELECT MAX(level) as level FROM ${this.escapeTableName(tableName)} WHERE descendant = ${parentId}`);
            return results && results[0] && results[0]["level"] ? parseInt(results[0]["level"]) + 1 : 1;
        });
    }
    /**
     * Replaces parameters in the given sql with special escaping character
     * and an array of parameter names to be passed to a query.
     */
    escapeQueryWithParameters(sql, parameters) {
        if (!parameters || !Object.keys(parameters).length)
            return [sql, []];
        const escapedParameters = [];
        const keys = Object.keys(parameters).map(parameter => "(:" + parameter + ")").join("|");
        sql = sql.replace(new RegExp(keys, "g"), (key) => {
            escapedParameters.push(parameters[key.substr(1)]);
            return "?";
        }); // todo: make replace only in value statements, otherwise problems
        return [sql, escapedParameters];
    }
    /**
     * Escapes a column name.
     */
    escapeColumnName(columnName) {
        return columnName; // "`" + columnName + "`";
    }
    /**
     * Escapes an alias.
     */
    escapeAliasName(aliasName) {
        return aliasName; // "`" + aliasName + "`";
    }
    /**
     * Escapes a table name.
     */
    escapeTableName(tableName) {
        return tableName; // "`" + tableName + "`";
    }
    /**
     * Prepares given value to a value to be persisted, based on its column type and metadata.
     */
    preparePersistentValue(value, column) {
        return ColumnTypes_1.ColumnTypes.preparePersistentValue(value, column.type);
    }
    /**
     * Prepares given value to a value to be persisted, based on its column type and metadata.
     */
    prepareHydratedValue(value, column) {
        return ColumnTypes_1.ColumnTypes.prepareHydratedValue(value, column.type);
    }
    /**
     * Access to the native implementation of the database.
     */
    nativeInterface() {
        return {
            driver: this.mysql,
            connection: this.databaseConnection ? this.databaseConnection.connection : undefined,
            pool: this.pool
        };
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    /**
     * If driver dependency is not given explicitly, then try to load it via "require".
     */
    loadDependencies() {
        if (!require)
            throw new DriverPackageLoadError_1.DriverPackageLoadError();
        try {
            this.mysql = require("mysql");
        }
        catch (e) {
            throw new DriverPackageNotInstalledError_1.DriverPackageNotInstalledError("Mysql", "mysql");
        }
    }
    /**
     * Parametrizes given object of values. Used to create column=value queries.
     */
    parametrize(objectLiteral) {
        return Object.keys(objectLiteral).map(key => this.escapeColumnName(key) + "=?");
    }
}
exports.MysqlDriver = MysqlDriver;

//# sourceMappingURL=MysqlDriver.js.map
