"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const ConnectionIsNotSetError_1 = require("./error/ConnectionIsNotSetError");
const PostgresSchemaBuilder_1 = require("../schema-builder/PostgresSchemaBuilder");
const DriverPackageNotInstalledError_1 = require("./error/DriverPackageNotInstalledError");
const DriverPackageLoadError_1 = require("./error/DriverPackageLoadError");
const DriverUtils_1 = require("./DriverUtils");
const ColumnTypes_1 = require("../metadata/types/ColumnTypes");
const TransactionAlreadyStartedError_1 = require("./error/TransactionAlreadyStartedError");
const TransactionNotStartedError_1 = require("./error/TransactionNotStartedError");
// todo(tests):
// check connection with url
// check if any of required option is not set exception to be thrown
//
/**
 * This driver organizes work with postgres database.
 */
class PostgresDriver {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(connectionOptions, logger, postgres) {
        /**
         * Pool of database connections.
         */
        this.databaseConnectionPool = [];
        this.options = DriverUtils_1.DriverUtils.buildDriverOptions(connectionOptions);
        this.logger = logger;
        this.postgres = postgres;
        // validate options to make sure everything is set
        DriverUtils_1.DriverUtils.validateDriverOptions(this.options);
        // if postgres package instance was not set explicitly then try to load it
        if (!postgres)
            this.loadDependencies();
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Performs connection to the database.
     * Based on pooling options, it can either create connection immediately,
     * either create a pool and create connection when needed.
     */
    connect() {
        // build connection options for the driver
        const options = Object.assign({}, {
            host: this.options.host,
            user: this.options.username,
            password: this.options.password,
            database: this.options.database,
            port: this.options.port
        }, this.options.extra || {});
        // pooling is enabled either when its set explicitly to true,
        // either when its not defined at all (e.g. enabled by default)
        if (this.options.usePool === undefined || this.options.usePool === true) {
            this.pool = new this.postgres.Pool(options);
            return Promise.resolve();
        }
        else {
            return new Promise((ok, fail) => {
                this.databaseConnection = {
                    id: 1,
                    connection: new this.postgres.Client(options),
                    isTransactionActive: false
                };
                this.databaseConnection.connection.connect((err) => err ? fail(err) : ok());
            });
        }
    }
    /**
     * Closes connection with database.
     */
    disconnect() {
        if (!this.databaseConnection && !this.pool)
            throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
        return new Promise((ok, fail) => {
            const handler = (err) => err ? fail(err) : ok();
            if (this.databaseConnection) {
                this.databaseConnection.connection.end(); // todo: check if it can emit errors
                this.databaseConnection = undefined;
            }
            if (this.databaseConnectionPool) {
                this.databaseConnectionPool.forEach(dbConnection => {
                    if (dbConnection && dbConnection.releaseCallback) {
                        dbConnection.releaseCallback();
                    }
                });
                this.pool.end(handler);
                this.pool = undefined;
                this.databaseConnectionPool = [];
            }
            ok();
        });
    }
    /**
     * Retrieves a new database connection.
     * If pooling is enabled then connection from the pool will be retrieved.
     * Otherwise active connection will be returned.
     */
    retrieveDatabaseConnection() {
        if (this.pool) {
            return new Promise((ok, fail) => {
                this.pool.connect((err, connection, release) => {
                    if (err) {
                        fail(err);
                        return;
                    }
                    let dbConnection = this.databaseConnectionPool.find(dbConnection => dbConnection.connection === connection);
                    if (!dbConnection) {
                        dbConnection = {
                            id: this.databaseConnectionPool.length,
                            connection: connection,
                            isTransactionActive: false
                        };
                        this.databaseConnectionPool.push(dbConnection);
                    }
                    dbConnection.releaseCallback = release;
                    ok(dbConnection);
                });
            });
        }
        if (this.databaseConnection)
            return Promise.resolve(this.databaseConnection);
        throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
    }
    /**
     * Releases database connection. This is needed when using connection pooling.
     * If connection is not from a pool, it should not be released.
     */
    releaseDatabaseConnection(dbConnection) {
        if (this.pool && dbConnection.releaseCallback) {
            dbConnection.releaseCallback();
            this.databaseConnectionPool.splice(this.databaseConnectionPool.indexOf(dbConnection), 1);
        }
        return Promise.resolve();
    }
    /**
     * Creates a schema builder which can be used to build database/table schemas.
     */
    createSchemaBuilder(dbConnection) {
        return new PostgresSchemaBuilder_1.PostgresSchemaBuilder(this, dbConnection);
    }
    /**
     * Removes all tables from the currently connected database.
     */
    clearDatabase(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
            const selectDropsQuery = `SELECT 'DROP TABLE IF EXISTS "' || tablename || '" CASCADE;' as query FROM pg_tables WHERE schemaname = 'public'`;
            const dropQueries = yield this.query(dbConnection, selectDropsQuery);
            yield Promise.all(dropQueries.map(q => this.query(dbConnection, q["query"])));
        });
    }
    /**
     * Starts transaction.
     */
    beginTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (dbConnection.isTransactionActive)
                throw new TransactionAlreadyStartedError_1.TransactionAlreadyStartedError();
            yield this.query(dbConnection, "START TRANSACTION");
            dbConnection.isTransactionActive = true;
        });
    }
    /**
     * Commits transaction.
     */
    commitTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new TransactionNotStartedError_1.TransactionNotStartedError();
            yield this.query(dbConnection, "COMMIT");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Rollbacks transaction.
     */
    rollbackTransaction(dbConnection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!dbConnection.isTransactionActive)
                throw new TransactionNotStartedError_1.TransactionNotStartedError();
            yield this.query(dbConnection, "ROLLBACK");
            dbConnection.isTransactionActive = false;
        });
    }
    /**
     * Executes a given SQL query.
     */
    query(dbConnection, query, parameters) {
        if (!this.databaseConnection && !this.pool)
            throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
        // console.log("query: ", query);
        // console.log("parameters: ", parameters);
        this.logger.logQuery(query);
        return new Promise((ok, fail) => {
            dbConnection.connection.query(query, parameters, (err, result) => {
                if (err) {
                    this.logger.logFailedQuery(query);
                    this.logger.logQueryError(err);
                    fail(err);
                }
                else {
                    ok(result.rows);
                }
            });
        });
    }
    /**
     * Insert a new row into given table.
     */
    insert(dbConnection, tableName, keyValues, idColumnName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
            const keys = Object.keys(keyValues);
            const columns = keys.map(key => this.escapeColumnName(key)).join(", ");
            const values = keys.map((key, index) => "$" + (index + 1)).join(",");
            const sql = `INSERT INTO ${this.escapeTableName(tableName)}(${columns}) VALUES (${values}) ${idColumnName ? " RETURNING " + this.escapeColumnName(idColumnName) : ""}`;
            const parameters = keys.map(key => keyValues[key]);
            const result = yield this.query(dbConnection, sql, parameters);
            if (idColumnName)
                return result[0][idColumnName];
            return result;
        });
    }
    /**
     * Updates rows that match given conditions in the given table.
     */
    update(dbConnection, tableName, valuesMap, conditions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
            const updateValues = this.parametrize(valuesMap).join(", ");
            const conditionString = this.parametrize(conditions, Object.keys(valuesMap).length).join(" AND ");
            const query = `UPDATE ${this.escapeTableName(tableName)} SET ${updateValues} ${conditionString ? (" WHERE " + conditionString) : ""}`;
            const updateParams = Object.keys(valuesMap).map(key => valuesMap[key]);
            const conditionParams = Object.keys(conditions).map(key => conditions[key]);
            const allParameters = updateParams.concat(conditionParams);
            yield this.query(dbConnection, query, allParameters);
        });
    }
    /**
     * Deletes from the given table by a given conditions.
     */
    delete(dbConnection, tableName, conditions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.databaseConnection && !this.pool)
                throw new ConnectionIsNotSetError_1.ConnectionIsNotSetError("postgres");
            const conditionString = this.parametrize(conditions).join(" AND ");
            const parameters = Object.keys(conditions).map(key => conditions[key]);
            const query = `DELETE FROM "${tableName}" WHERE ${conditionString}`;
            yield this.query(dbConnection, query, parameters);
        });
    }
    /**
     * Inserts rows into closure table.
     */
    insertIntoClosureTable(dbConnection, tableName, newEntityId, parentId, hasLevel) {
        return __awaiter(this, void 0, void 0, function* () {
            let sql = "";
            if (hasLevel) {
                sql = `INSERT INTO ${this.escapeTableName(tableName)}(ancestor, descendant, level) ` +
                    `SELECT ancestor, ${newEntityId}, level + 1 FROM ${this.escapeTableName(tableName)} WHERE descendant = ${parentId} ` +
                    `UNION ALL SELECT ${newEntityId}, ${newEntityId}, 1`;
            }
            else {
                sql = `INSERT INTO ${this.escapeTableName(tableName)}(ancestor, descendant) ` +
                    `SELECT ancestor, ${newEntityId} FROM ${this.escapeTableName(tableName)} WHERE descendant = ${parentId} ` +
                    `UNION ALL SELECT ${newEntityId}, ${newEntityId}`;
            }
            yield this.query(dbConnection, sql);
            const results = yield this.query(dbConnection, `SELECT MAX(level) as level FROM ${tableName} WHERE descendant = ${parentId}`);
            return results && results[0] && results[0]["level"] ? parseInt(results[0]["level"]) + 1 : 1;
        });
    }
    /**
     * Prepares given value to a value to be persisted, based on its column type and metadata.
     */
    preparePersistentValue(value, column) {
        return ColumnTypes_1.ColumnTypes.preparePersistentValue(value, column.type);
    }
    /**
     * Prepares given value to a value to be persisted, based on its column type and metadata.
     */
    prepareHydratedValue(value, column) {
        return ColumnTypes_1.ColumnTypes.prepareHydratedValue(value, column.type);
    }
    /**
     * Replaces parameters in the given sql with special escaping character
     * and an array of parameter names to be passed to a query.
     */
    escapeQueryWithParameters(sql, parameters) {
        if (!parameters || !Object.keys(parameters).length)
            return [sql, []];
        const builtParameters = [];
        const keys = Object.keys(parameters).map(parameter => "(:" + parameter + ")").join("|");
        sql = sql.replace(new RegExp(keys, "g"), (key) => {
            const value = parameters[key.substr(1)];
            if (value instanceof Array) {
                return value.map((v) => {
                    builtParameters.push(v);
                    return "$" + builtParameters.length;
                });
            }
            else {
                builtParameters.push(value);
            }
            return "$" + builtParameters.length;
        }); // todo: make replace only in value statements, otherwise problems
        return [sql, builtParameters];
    }
    /**
     * Escapes a column name.
     */
    escapeColumnName(columnName) {
        return "\"" + columnName + "\"";
    }
    /**
     * Escapes an alias.
     */
    escapeAliasName(aliasName) {
        return "\"" + aliasName + "\"";
    }
    /**
     * Escapes a table name.
     */
    escapeTableName(tableName) {
        return "\"" + tableName + "\"";
    }
    /**
     * Access to the native implementation of the database.
     */
    nativeInterface() {
        return {
            driver: this.postgres,
            connection: this.databaseConnection ? this.databaseConnection.connection : undefined,
            pool: this.pool
        };
    }
    // -------------------------------------------------------------------------
    // Protected Methods
    // -------------------------------------------------------------------------
    /**
     * If driver dependency is not given explicitly, then try to load it via "require".
     */
    loadDependencies() {
        if (!require)
            throw new DriverPackageLoadError_1.DriverPackageLoadError();
        try {
            this.postgres = require("pg");
        }
        catch (e) {
            throw new DriverPackageNotInstalledError_1.DriverPackageNotInstalledError("Postgres", "pg");
        }
    }
    /**
     * Parametrizes given object of values. Used to create column=value queries.
     */
    parametrize(objectLiteral, startIndex = 0) {
        return Object.keys(objectLiteral).map((key, index) => this.escapeColumnName(key) + "=$" + (startIndex + index + 1));
    }
}
exports.PostgresDriver = PostgresDriver;

//# sourceMappingURL=PostgresDriver.js.map
