"use strict";
const TargetMetadataArgsCollection_1 = require("./collection/TargetMetadataArgsCollection");
const PropertyMetadataArgsCollection_1 = require("./collection/PropertyMetadataArgsCollection");
/**
 * Storage all metadatas of all available types: tables, fields, subscribers, relations, etc.
 * Each metadata represents some specifications of what it represents.
 */
class MetadataArgsStorage {
    constructor() {
        // todo: type in function validation, inverse side function validation
        // todo: check on build for duplicate names, since naming checking was removed from MetadataStorage
        // todo: duplicate name checking for: table, relation, column, index, naming strategy, join tables/columns?
        // todo: check for duplicate targets too since this check has been removed too
        // -------------------------------------------------------------------------
        // Properties
        // -------------------------------------------------------------------------
        this.tables = new TargetMetadataArgsCollection_1.TargetMetadataArgsCollection();
        this.namingStrategies = new TargetMetadataArgsCollection_1.TargetMetadataArgsCollection();
        this.entitySubscribers = new TargetMetadataArgsCollection_1.TargetMetadataArgsCollection();
        this.indices = new TargetMetadataArgsCollection_1.TargetMetadataArgsCollection();
        this.columns = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.relations = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.joinColumns = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.joinTables = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.entityListeners = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.relationCounts = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.relationIds = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
        this.embeddeds = new PropertyMetadataArgsCollection_1.PropertyMetadataArgsCollection();
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Gets merged (with all abstract classes) table metadatas for the given classes.
     */
    getMergedTableMetadatas(classes) {
        const allTableMetadataArgs = classes ? this.tables.filterByTargets(classes) : this.tables;
        const tableMetadatas = allTableMetadataArgs.filter(table => table.type === "regular" || table.type === "closure");
        return tableMetadatas.map(tableMetadata => {
            return this.mergeWithAbstract(allTableMetadataArgs, tableMetadata);
        });
    }
    /**
     * Gets merged (with all abstract classes) embeddable table metadatas for the given classes.
     */
    getMergedEmbeddableTableMetadatas(classes) {
        const tables = classes ? this.tables.filterByTargets(classes) : this.tables;
        const embeddableTableMetadatas = tables.filter(table => table.type === "embeddable");
        return embeddableTableMetadatas.map(embeddableTableMetadata => {
            return this.mergeWithEmbeddable(embeddableTableMetadatas, embeddableTableMetadata);
        });
    }
    // -------------------------------------------------------------------------
    // Private Methods
    // -------------------------------------------------------------------------
    /**
     */
    mergeWithAbstract(allTableMetadatas, tableMetadata) {
        const indices = this.indices.filterByTarget(tableMetadata.target);
        const columns = this.columns.filterByTarget(tableMetadata.target);
        const relations = this.relations.filterByTarget(tableMetadata.target);
        const joinColumns = this.joinColumns.filterByTarget(tableMetadata.target);
        const joinTables = this.joinTables.filterByTarget(tableMetadata.target);
        const entityListeners = this.entityListeners.filterByTarget(tableMetadata.target);
        const relationCounts = this.relationCounts.filterByTarget(tableMetadata.target);
        const relationIds = this.relationIds.filterByTarget(tableMetadata.target);
        const embeddeds = this.embeddeds.filterByTarget(tableMetadata.target);
        allTableMetadatas
            .filter(metadata => {
            if (!tableMetadata.target || !metadata.target)
                return false;
            if (!(tableMetadata.target instanceof Function) || !(metadata.target instanceof Function))
                return false;
            return this.isInherited(tableMetadata.target, metadata.target); // todo: fix it
        })
            .forEach(parentMetadata => {
            const metadatasFromAbstract = this.mergeWithAbstract(allTableMetadatas, parentMetadata);
            metadatasFromAbstract.columns
                .filterRepeatedMetadatas(columns)
                .forEach(metadata => columns.push(metadata));
            metadatasFromAbstract.relations
                .filterRepeatedMetadatas(relations)
                .forEach(metadata => relations.push(metadata));
            metadatasFromAbstract.joinColumns
                .filterRepeatedMetadatas(joinColumns)
                .forEach(metadata => joinColumns.push(metadata));
            metadatasFromAbstract.joinTables
                .filterRepeatedMetadatas(joinTables)
                .forEach(metadata => joinTables.push(metadata));
            metadatasFromAbstract.entityListeners
                .filterRepeatedMetadatas(entityListeners)
                .forEach(metadata => entityListeners.push(metadata));
            metadatasFromAbstract.relationCounts
                .filterRepeatedMetadatas(relationCounts)
                .forEach(metadata => relationCounts.push(metadata));
            metadatasFromAbstract.relationIds
                .filterRepeatedMetadatas(relationIds)
                .forEach(metadata => relationIds.push(metadata));
            metadatasFromAbstract.embeddeds
                .filterRepeatedMetadatas(embeddeds)
                .forEach(metadata => embeddeds.push(metadata));
        });
        return {
            table: tableMetadata,
            indices: indices,
            columns: columns,
            relations: relations,
            joinColumns: joinColumns,
            joinTables: joinTables,
            entityListeners: entityListeners,
            relationCounts: relationCounts,
            relationIds: relationIds,
            embeddeds: embeddeds
        };
    }
    /**
     */
    mergeWithEmbeddable(allTableMetadatas, tableMetadata) {
        const columns = this.columns.filterByTarget(tableMetadata.target);
        allTableMetadatas
            .filter(metadata => {
            if (!tableMetadata.target || !metadata.target)
                return false;
            if (!(tableMetadata.target instanceof Function) || !(metadata.target instanceof Function))
                return false;
            return this.isInherited(tableMetadata.target, metadata.target); // todo: fix it for entity schema
        })
            .forEach(parentMetadata => {
            const metadatasFromParents = this.mergeWithEmbeddable(allTableMetadatas, parentMetadata);
            metadatasFromParents.columns
                .filterRepeatedMetadatas(columns)
                .forEach(metadata => columns.push(metadata));
        });
        return {
            table: tableMetadata,
            columns: columns
        };
    }
    /**
     * Checks if this table is inherited from another table.
     */
    isInherited(target1, target2) {
        // we cannot use instanceOf in this method, because we need order of inherited tables, to ensure that
        // properties get inherited in a right order. To achieve it we can only check a first parent of the class
        // return this.target.prototype instanceof anotherTable.target;
        return Object.getPrototypeOf(target1.prototype).constructor === target2;
    }
}
exports.MetadataArgsStorage = MetadataArgsStorage;

//# sourceMappingURL=MetadataArgsStorage.js.map
