"use strict";
const EntityMetadata_1 = require("../metadata/EntityMetadata");
const ColumnMetadata_1 = require("../metadata/ColumnMetadata");
const ForeignKeyMetadata_1 = require("../metadata/ForeignKeyMetadata");
const EntityMetadataValidator_1 = require("./EntityMetadataValidator");
const IndexMetadata_1 = require("../metadata/IndexMetadata");
const JoinColumnMetadata_1 = require("../metadata/JoinColumnMetadata");
const TableMetadata_1 = require("../metadata/TableMetadata");
const index_1 = require("../index");
const RelationMetadata_1 = require("../metadata/RelationMetadata");
const JoinTableMetadata_1 = require("../metadata/JoinTableMetadata");
const JunctionEntityMetadataBuilder_1 = require("./JunctionEntityMetadataBuilder");
const ClosureJunctionEntityMetadataBuilder_1 = require("./ClosureJunctionEntityMetadataBuilder");
const EmbeddedMetadata_1 = require("../metadata/EmbeddedMetadata");
const MetadataArgsStorage_1 = require("../metadata-args/MetadataArgsStorage");
/**
 * Aggregates all metadata: table, column, relation into one collection grouped by tables for a given set of classes.
 *
 * @internal
 */
class EntityMetadataBuilder {
    // todo: type in function validation, inverse side function validation
    // todo: check on build for duplicate names, since naming checking was removed from MetadataStorage
    // todo: duplicate name checking for: table, relation, column, index, naming strategy, join tables/columns?
    // todo: check if multiple tree parent metadatas in validator
    // todo: tree decorators can be used only on closure table (validation)
    // todo: throw error if parent tree metadata was not specified in a closure table
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    buildFromSchemas(lazyRelationsWrapper, namingStrategy, schemas) {
        const metadataArgsStorage = new MetadataArgsStorage_1.MetadataArgsStorage();
        schemas.forEach(schema => {
            // add table metadata args from the schema
            const tableSchema = schema.table || {};
            const table = {
                target: schema.target || schema.name,
                name: tableSchema.name,
                type: tableSchema.type || "regular",
                // targetId: schema.name,
                orderBy: tableSchema.orderBy,
                primaryKeys: tableSchema.primaryKeys
            };
            metadataArgsStorage.tables.add(table);
            // add columns metadata args from the schema
            Object.keys(schema.columns).forEach(columnName => {
                const columnSchema = schema.columns[columnName];
                let mode = "regular";
                if (columnSchema.primary)
                    mode = "primary";
                if (columnSchema.createDate)
                    mode = "createDate";
                if (columnSchema.updateDate)
                    mode = "updateDate";
                if (columnSchema.version)
                    mode = "version";
                if (columnSchema.treeChildrenCount)
                    mode = "treeChildrenCount";
                if (columnSchema.treeLevel)
                    mode = "treeLevel";
                const column = {
                    target: schema.target || schema.name,
                    mode: mode,
                    propertyName: columnName,
                    // todo: what to do with it?: propertyType: 
                    options: {
                        type: columnSchema.type,
                        name: columnSchema.name,
                        length: columnSchema.length,
                        generated: columnSchema.generated,
                        unique: columnSchema.unique,
                        nullable: columnSchema.nullable,
                        columnDefinition: columnSchema.columnDefinition,
                        comment: columnSchema.comment,
                        oldColumnName: columnSchema.oldColumnName,
                        precision: columnSchema.precision,
                        scale: columnSchema.scale,
                        collation: columnSchema.collation
                    }
                };
                metadataArgsStorage.columns.add(column);
            });
            // add relation metadata args from the schema
            if (schema.relations) {
                Object.keys(schema.relations).forEach(relationName => {
                    const relationSchema = schema.relations[relationName];
                    const relation = {
                        target: schema.target || schema.name,
                        propertyName: relationName,
                        relationType: relationSchema.type,
                        isLazy: relationSchema.isLazy || false,
                        type: relationSchema.target,
                        inverseSideProperty: relationSchema.inverseSide,
                        isTreeParent: relationSchema.isTreeParent,
                        isTreeChildren: relationSchema.isTreeChildren,
                        options: {
                            cascadeAll: relationSchema.cascadeAll,
                            cascadeInsert: relationSchema.cascadeInsert,
                            cascadeUpdate: relationSchema.cascadeUpdate,
                            cascadeRemove: relationSchema.cascadeRemove,
                            oldColumnName: relationSchema.oldColumnName,
                            nullable: relationSchema.nullable,
                            onDelete: relationSchema.onDelete
                        }
                    };
                    metadataArgsStorage.relations.add(relation);
                    // add join column
                    if (relationSchema.joinColumn) {
                        if (typeof relationSchema.joinColumn === "boolean") {
                            const joinColumn = {
                                target: schema.target || schema.name,
                                propertyName: relationName
                            };
                            metadataArgsStorage.joinColumns.push(joinColumn);
                        }
                        else {
                            const joinColumn = {
                                target: schema.target || schema.name,
                                propertyName: relationName,
                                name: relationSchema.joinColumn.name,
                                referencedColumnName: relationSchema.joinColumn.referencedColumnName
                            };
                            metadataArgsStorage.joinColumns.push(joinColumn);
                        }
                    }
                    // add join table
                    if (relationSchema.joinTable) {
                        if (typeof relationSchema.joinTable === "boolean") {
                            const joinTable = {
                                target: schema.target || schema.name,
                                propertyName: relationName
                            };
                            metadataArgsStorage.joinTables.push(joinTable);
                        }
                        else {
                            const joinTable = {
                                target: schema.target || schema.name,
                                propertyName: relationName,
                                name: relationSchema.joinTable.name,
                                joinColumn: relationSchema.joinTable.joinColumn,
                                inverseJoinColumn: relationSchema.joinTable.inverseJoinColumn
                            };
                            metadataArgsStorage.joinTables.push(joinTable);
                        }
                    }
                });
            }
        });
        return this.build(lazyRelationsWrapper, metadataArgsStorage, namingStrategy);
    }
    /**
     * Builds a complete metadata aggregations for the given entity classes.
     */
    buildFromMetadataArgsStorage(lazyRelationsWrapper, namingStrategy, entityClasses) {
        return this.build(lazyRelationsWrapper, index_1.getMetadataArgsStorage(), namingStrategy, entityClasses);
    }
    build(lazyRelationsWrapper, metadataArgsStorage, namingStrategy, entityClasses) {
        const embeddableMergedArgs = metadataArgsStorage.getMergedEmbeddableTableMetadatas(entityClasses);
        const entityMetadatas = metadataArgsStorage.getMergedTableMetadatas(entityClasses).map(mergedArgs => {
            // find embeddable tables for embeddeds registered in this table and create EmbeddedMetadatas from them
            const embeddeds = [];
            mergedArgs.embeddeds.forEach(embedded => {
                const embeddableTable = embeddableMergedArgs.find(mergedArgs => mergedArgs.table.target === embedded.type());
                if (embeddableTable) {
                    const table = new TableMetadata_1.TableMetadata(embeddableTable.table);
                    const columns = embeddableTable.columns.map(args => new ColumnMetadata_1.ColumnMetadata(args));
                    embeddeds.push(new EmbeddedMetadata_1.EmbeddedMetadata(embedded.type(), embedded.propertyName, table, columns));
                }
            });
            // create metadatas from args
            const table = new TableMetadata_1.TableMetadata(mergedArgs.table);
            const columns = mergedArgs.columns.map(args => new ColumnMetadata_1.ColumnMetadata(args));
            const relations = mergedArgs.relations.map(args => new RelationMetadata_1.RelationMetadata(args));
            const indices = mergedArgs.indices.map(args => new IndexMetadata_1.IndexMetadata(args));
            // create a new entity metadata
            const entityMetadata = new EntityMetadata_1.EntityMetadata(lazyRelationsWrapper, {
                namingStrategy: namingStrategy,
                tableMetadata: table,
                columnMetadatas: columns,
                relationMetadatas: relations,
                indexMetadatas: indices,
                embeddedMetadatas: embeddeds
            });
            // create entity's relations join tables
            entityMetadata.manyToManyRelations.forEach(relation => {
                const joinTableMetadata = mergedArgs.joinTables.findByProperty(relation.propertyName);
                if (joinTableMetadata) {
                    const joinTable = new JoinTableMetadata_1.JoinTableMetadata(joinTableMetadata);
                    relation.joinTable = joinTable;
                    joinTable.relation = relation;
                }
            });
            // create entity's relations join columns
            entityMetadata.oneToOneRelations
                .concat(entityMetadata.manyToOneRelations)
                .forEach(relation => {
                // since for many-to-one relations having JoinColumn is not required on decorators level, we need to go
                // throw all of them which don't have JoinColumn decorators and create it for them
                let joinColumnMetadata = mergedArgs.joinColumns.findByProperty(relation.propertyName);
                if (!joinColumnMetadata && relation.isManyToOne) {
                    joinColumnMetadata = {
                        target: relation.target,
                        propertyName: relation.propertyName
                    };
                }
                if (joinColumnMetadata) {
                    const joinColumn = new JoinColumnMetadata_1.JoinColumnMetadata(joinColumnMetadata);
                    relation.joinColumn = joinColumn;
                    joinColumn.relation = relation;
                }
            });
            // save relation id-s data
            entityMetadata.relations.forEach(relation => {
                const relationIdMetadata = mergedArgs.relationIds.find(relationId => {
                    if (relationId.relation instanceof Function)
                        return relation.propertyName === relationId.relation(entityMetadata.createPropertiesMap());
                    return relation.propertyName === relationId.relation;
                });
                if (relationIdMetadata) {
                    if (relation.isOneToOneNotOwner || relation.isOneToMany)
                        throw new Error(`RelationId cannot be used for the one-to-one without join column or one-to-many relations.`);
                    relation.idField = relationIdMetadata.propertyName;
                }
            });
            // save relation counter-s data
            entityMetadata.relations.forEach(relation => {
                const relationCountMetadata = mergedArgs.relationCounts.find(relationCount => {
                    if (relationCount.relation instanceof Function)
                        return relation.propertyName === relationCount.relation(entityMetadata.createPropertiesMap());
                    return relation.propertyName === relationCount.relation;
                });
                if (relationCountMetadata)
                    relation.countField = relationCountMetadata.propertyName;
            });
            // add lazy initializer for entity relations
            if (entityMetadata.target instanceof Function) {
                entityMetadata.relations
                    .filter(relation => relation.isLazy)
                    .forEach(relation => {
                    lazyRelationsWrapper.wrap(entityMetadata.target.prototype, relation);
                });
            }
            return entityMetadata;
        });
        // after all metadatas created we set inverse side (related) entity metadatas for all relation metadatas
        entityMetadatas.forEach(entityMetadata => {
            entityMetadata.relations.forEach(relation => {
                const inverseEntityMetadata = entityMetadatas.find(m => m.target === relation.type || (typeof relation.type === "string" && m.targetName === relation.type));
                if (!inverseEntityMetadata)
                    throw new Error("Entity metadata for " + entityMetadata.name + "#" + relation.propertyName + " was not found.");
                relation.inverseEntityMetadata = inverseEntityMetadata;
            });
        });
        // check for errors in a built metadata schema (we need to check after relationEntityMetadata is set)
        index_1.getFromContainer(EntityMetadataValidator_1.EntityMetadataValidator).validateMany(entityMetadatas);
        // generate columns and foreign keys for tables with relations
        entityMetadatas.forEach(metadata => {
            metadata.relationsWithJoinColumns.forEach(relation => {
                // find relational column and if it does not exist - add it
                const inverseSideColumn = relation.joinColumn.referencedColumn;
                let relationalColumn = metadata.columns.find(column => column.name === relation.name);
                if (!relationalColumn) {
                    relationalColumn = new ColumnMetadata_1.ColumnMetadata({
                        target: metadata.target,
                        propertyName: relation.name,
                        propertyType: inverseSideColumn.propertyType,
                        mode: "virtual",
                        options: {
                            type: inverseSideColumn.type,
                            oldColumnName: relation.oldColumnName,
                            nullable: relation.isNullable
                        }
                    });
                    metadata.addColumn(relationalColumn);
                }
                // create and add foreign key
                const foreignKey = new ForeignKeyMetadata_1.ForeignKeyMetadata([relationalColumn], relation.inverseEntityMetadata.table, [inverseSideColumn], relation.onDelete);
                foreignKey.entityMetadata = metadata;
                metadata.foreignKeys.push(foreignKey);
            });
        });
        // generate junction tables for all closure tables
        entityMetadatas.forEach(metadata => {
            if (!metadata.table.isClosure)
                return;
            const closureJunctionEntityMetadata = index_1.getFromContainer(ClosureJunctionEntityMetadataBuilder_1.ClosureJunctionEntityMetadataBuilder).build(lazyRelationsWrapper, {
                namingStrategy: namingStrategy,
                table: metadata.table,
                primaryColumn: metadata.primaryColumn,
                hasTreeLevelColumn: metadata.hasTreeLevelColumn
            });
            metadata.closureJunctionTable = closureJunctionEntityMetadata;
            entityMetadatas.push(closureJunctionEntityMetadata);
        });
        // generate junction tables for all many-to-many tables
        entityMetadatas.forEach(metadata => {
            metadata.ownerManyToManyRelations.forEach(relation => {
                const junctionEntityMetadata = index_1.getFromContainer(JunctionEntityMetadataBuilder_1.JunctionEntityMetadataBuilder).build(lazyRelationsWrapper, {
                    namingStrategy: namingStrategy,
                    firstTable: metadata.table,
                    secondTable: relation.inverseEntityMetadata.table,
                    joinTable: relation.joinTable
                });
                relation.junctionEntityMetadata = junctionEntityMetadata;
                if (relation.hasInverseSide)
                    relation.inverseRelation.junctionEntityMetadata = junctionEntityMetadata;
                entityMetadatas.push(junctionEntityMetadata);
            });
        });
        return entityMetadatas;
    }
}
exports.EntityMetadataBuilder = EntityMetadataBuilder;

//# sourceMappingURL=EntityMetadataBuilder.js.map
