import { PropertyMetadata } from "./PropertyMetadata";
import { ColumnMetadataArgs } from "../metadata-args/ColumnMetadataArgs";
import { ColumnType } from "./types/ColumnTypes";
import { EntityMetadata } from "./EntityMetadata";
import { EmbeddedMetadata } from "./EmbeddedMetadata";
/**
 * Kinda type of the column. Not a type in the database, but locally used type to determine what kind of column
 * we are working with.
 * For example, "primary" means that it will be a primary column, or "createDate" means that it will create a create
 * date column.
 */
export declare type ColumnMode = "regular" | "virtual" | "primary" | "createDate" | "updateDate" | "version" | "treeChildrenCount" | "treeLevel";
/**
 * This metadata contains all information about entity's column.
 */
export declare class ColumnMetadata extends PropertyMetadata {
    /**
     * Entity metadata where this column metadata is.
     */
    entityMetadata: EntityMetadata;
    /**
     * Embedded metadata where this column metadata is.
     */
    embeddedMetadata: EmbeddedMetadata;
    /**
     * The real reflected property type.
     */
    readonly propertyType: string;
    /**
     * The database type of the column.
     */
    readonly type: ColumnType;
    /**
     * Column's mode in which this column is working.
     */
    readonly mode: ColumnMode;
    /**
     * Type's length in the database.
     */
    readonly length: string;
    /**
     * Indicates if this column is generated (auto increment or generated other way).
     */
    readonly isGenerated: boolean;
    /**
     * Indicates if value in the database should be unique or not.
     */
    readonly isUnique: boolean;
    /**
     * Indicates if column can contain nulls or not.
     */
    readonly isNullable: boolean;
    /**
     * Extra sql definition for the given column.
     * Can be used to make temporary tweaks. Not recommended to use.
     */
    readonly columnDefinition: string;
    /**
     * Column comment.
     */
    readonly comment: string;
    /**
     * Old column name. Used to correctly alter tables when column name is changed.
     * Can be used to make temporary tweaks. Not recommended to use.
     */
    readonly oldColumnName: string;
    /**
     * The precision for a decimal (exact numeric) column (applies only for decimal column), which is the maximum
     * number of digits that are stored for the values.
     */
    readonly precision: number;
    /**
     * The scale for a decimal (exact numeric) column (applies only for decimal column), which represents the number
     * of digits to the right of the decimal point and must not be greater than precision.
     */
    readonly scale: number;
    /**
     * Column collation. Note that not all databases support it.
     */
    readonly collation: string;
    /**
     * Indicates if this date column will contain a timezone.
     * Used only for date-typed column types.
     * Note that timezone option is not supported by all databases (only postgres for now).
     */
    readonly timezone: boolean;
    /**
     * Column name to be used in the database.
     */
    private _name;
    constructor(args: ColumnMetadataArgs);
    /**
     * Column name in the database.
     */
    readonly name: string;
    readonly target: string | Function;
    /**
     * Indicates if this column is in embedded, not directly in the table.
     */
    readonly isInEmbedded: boolean;
    /**
     * Indicates if this column is a primary key.
     */
    readonly isPrimary: boolean;
    /**
     * Indicates if column is virtual. Virtual columns are not mapped to the entity.
     */
    readonly isVirtual: boolean;
    /**
     * Indicates if this column contains an entity creation date.
     */
    readonly isCreateDate: boolean;
    /**
     * Indicates if this column contains an entity update date.
     */
    readonly isUpdateDate: boolean;
    /**
     * Indicates if this column contains an entity version.
     */
    readonly isVersion: boolean;
    /**
     * Gets embedded property in which column is.
     */
    readonly embeddedProperty: string;
    hasEntityValue(entity: any): boolean;
    getEntityValue(entity: any): any;
    readonly normalizedDataType: any;
}
