"use strict";
const Rx = require("rxjs/Rx");
/**
 * Repository is supposed to work with your entity objects. Find entities, insert, update, delete, etc.
 * This version of Repository is using rxjs library and Observables instead of promises.
 *
 * @experimental
 */
class ReactiveRepository {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(repository) {
        this.repository = repository;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Repository's target is an object (or unique string in the case if entity is loaded from a schema) that is managed
     * by this repository.
     */
    get target() {
        return this.repository.target;
    }
    /**
     * Checks if entity has an id.
     */
    hasId(entity) {
        return this.repository.hasId(entity);
    }
    /**
     * Creates a new query builder that can be used to build a sql query.
     */
    createQueryBuilder(alias) {
        return this.repository.createQueryBuilder(alias);
    }
    /**
     * Creates a new entity instance or instances.
     * Can copy properties from the given object into new entities.
     */
    create(plainObjectOrObjects) {
        if (plainObjectOrObjects instanceof Array) {
            return this.repository.create(plainObjectOrObjects);
        }
        else if (plainObjectOrObjects) {
            return this.repository.create(plainObjectOrObjects);
        }
        else {
            return this.repository.create();
        }
    }
    /**
     * Creates a new entity from the given plan javascript object. If entity already exist in the database, then
     * it loads it (and everything related to it), replaces all values with the new ones from the given object
     * and returns this new entity. This new entity is actually a loaded from the db entity with all properties
     * replaced from the new object.
     */
    preload(object) {
        return Rx.Observable.fromPromise(this.repository.preload(object));
    }
    /**
     * Merges two entities into one new entity.
     */
    merge(...objects) {
        return this.repository.merge(...objects);
    }
    /**
     * Persists (saves) a given entity in the database. If entity does not exist in the database then it inserts it,
     * else if entity already exist in the database then it updates it.
     */
    persist(entity) {
        return Rx.Observable.fromPromise(this.repository.persist(entity));
    }
    /**
     * Removes a given entity from the database.
     */
    remove(entity) {
        return Rx.Observable.fromPromise(this.repository.remove(entity));
    }
    /**
     * Finds entities that match given conditions.
     */
    find(conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return Rx.Observable.fromPromise(this.repository.find(conditionsOrFindOptions, options));
        }
        else if (conditionsOrFindOptions) {
            return Rx.Observable.fromPromise(this.repository.find(conditionsOrFindOptions));
        }
        else {
            return Rx.Observable.fromPromise(this.repository.find());
        }
    }
    /**
     * Finds entities that match given conditions.
     */
    findAndCount(conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return Rx.Observable.fromPromise(this.repository.findAndCount(conditionsOrFindOptions, options));
        }
        else if (conditionsOrFindOptions) {
            return Rx.Observable.fromPromise(this.repository.findAndCount(conditionsOrFindOptions));
        }
        else {
            return Rx.Observable.fromPromise(this.repository.findAndCount());
        }
    }
    /**
     * Finds first entity that matches given conditions.
     */
    findOne(conditionsOrFindOptions, options) {
        if (conditionsOrFindOptions && options) {
            return Rx.Observable.fromPromise(this.repository.findOne(conditionsOrFindOptions, options));
        }
        else if (conditionsOrFindOptions) {
            return Rx.Observable.fromPromise(this.repository.findOne(conditionsOrFindOptions));
        }
        else {
            return Rx.Observable.fromPromise(this.repository.findOne());
        }
    }
    /**
     * Finds entity with given id.
     */
    findOneById(id, options) {
        return Rx.Observable.fromPromise(this.repository.findOneById(id, options));
    }
    /**
     * Executes raw SQL query and returns raw database results.
     */
    query(query) {
        return Rx.Observable.fromPromise(this.repository.query(query));
    }
    /**
     * Wraps given function execution (and all operations made there) in a transaction.
     */
    transaction(runInTransaction) {
        return Rx.Observable.fromPromise(this.repository.transaction(runInTransaction));
    }
}
exports.ReactiveRepository = ReactiveRepository;

//# sourceMappingURL=ReactiveRepository.js.map
